/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2021 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.12.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#include "system_gd32vf103.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/

#define OS_TIMER_FREQ    (SystemCoreClock / 4)
#define OS_TICK_FREQ     (1000u)
#define OS_INT_FREQ      (OS_TICK_FREQ)

#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/

#define ECLIC_EXCEPTION     (0x7FFu)

#define NUM_INTERRUPTS      (32)

//
//  MTIME
//
#define MTIME_BASE_ADDR     (0xD1000000u)
#define MTIME_COUNTER_LOW   (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0x000u))
#define MTIME_COUNTER_HIGH  (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0x004u))
#define MTIME_COMPARE_LOW   (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0x008u))
#define MTIME_COMPARE_HIGH  (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0x00Cu))
#define MTIME_MSTOP         (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0xFF8u))
#define MTIME_MSIP          (*(volatile OS_U32*)(MTIME_BASE_ADDR + 0xFFCu))  // Timer can generate a software interrupt by writing 1 in the LSB

//
// ECLIC
//
#define ECLIC_BASE_ADDR     (0xD2000000u)

//
// GPIOC
//
#define GPIOC_BASE_ADDR     (0x40011000u)
#define GPIOC_CTL0          (*(volatile unsigned long*)(GPIOC_BASE_ADDR + 0x00u))  // Port control register 0
#define GPIOC_OCTL          (*(volatile unsigned long*)(GPIOC_BASE_ADDR + 0x0Cu))  // Port output control register
#define GPIOC_BOP           (*(volatile unsigned long*)(GPIOC_BASE_ADDR + 0x10u))  // Port bit operate register

//
// Reset and Clock Unit (RCU)
//
#define RCU_BASE_ADDR       (0x40021000u)
#define RCU_APB1EN          (*(volatile unsigned long*)(RCU_BASE_ADDR + 0x1Cu))

//
// RTC
//
#define RTC_BASE_ADDR       (0x40002800u)
#define RTC_INTEN           (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x00u)
#define RTC_CTL             (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x04u)
#define RTC_PSCH            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x08u)
#define RTC_PSCL            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x0Cu)
#define RTC_DIVH            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x10u)
#define RTC_DIVL            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x14u)
#define RTC_CNTH            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x18u)
#define RTC_CNTL            (*(volatile unsigned long*)(RTC_BASE_ADDR + 0x1Cu)

/*********************************************************************
*
*       Externals
*
**********************************************************************
*/

extern void* vectors;
extern void* trap_entry;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_U32 _OS_RISCV_HartID = 0u;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _MTIME_GetCounter()
*
*  Function description
*    Returns the current mtime hardware timer count value.
*
*  Return value
*    Current mtime timer count value.
*/
static OS_U64 _MTIME_GetCounter(void) {
  OS_U64 Counter;
  OS_U32 High;
  OS_U32 Low;

  do {
    High = MTIME_COUNTER_HIGH;           // Read high 32 bit
    Low  = MTIME_COUNTER_LOW;            // Read low 32 bit
  } while (High != MTIME_COUNTER_HIGH);  // If High 32 bit changed, low 32bit might not match high 32bit. Thus read timer again.
  Counter = (((OS_U64)High << 32) + (OS_U64)Low);
  return Counter;
}

/*********************************************************************
*
*       _MTIME_SetCounter()
*
*  Function description
*    Sets the current hardware timer count value.
*/
static void _MTIME_SetCounter(OS_U64 Value) {
  MTIME_COUNTER_HIGH = (OS_U32)(Value >> 32);
  MTIME_COUNTER_LOW  = (OS_U32)(Value);
}

/*********************************************************************
*
*       _MTIME_GetCompare()
*
*  Function description
*    Returns the current mtime hardware timer compare value.
*
*  Return value
*    Current mtime timer compare value.
*/
static OS_U64 _MTIME_GetCompare(void) {
  OS_U64 Compare;
  OS_U32 High;
  OS_U32 Low;

  High    = MTIME_COMPARE_HIGH;  // Read high 32 bit
  Low     = MTIME_COMPARE_LOW;   // Read low 32 bit
  Compare = (((OS_U64)High << 32) + (OS_U64)Low);
  return Compare;
}

/*********************************************************************
*
*       _MTIME_SetCompare()
*
*  Function description
*    Sets the current hardware timer compare value.
*/
static void _MTIME_SetCompare(OS_U64 Value) {
  MTIME_COMPARE_HIGH = (OS_U32)(Value >> 32);
  MTIME_COMPARE_LOW  = (OS_U32)(Value);
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  unsigned int Result;
  OS_I64       Diff;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = _MTIME_GetCompare();
  Counter = _MTIME_GetCounter();
  Diff    = (OS_I64)(Counter - Compare);
  if (Diff < 0) {
    Result = (unsigned int)(OS_TIMER_RELOAD + Diff);
  }  else {
    Result = (unsigned int)(Diff);
  }
  return Result;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  unsigned int IsPending;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = _MTIME_GetCompare();
  Counter = _MTIME_GetCounter();
  if (Compare <= Counter) {
    IsPending = 1u;
  } else {
    IsPending = 0u;
  }
  return IsPending;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_ISR_Software()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    OS_ISR_Software() is called when the Machine Software interrupt is pending.
*    Machine Software interrupt becomes pending when (COREPLEX_MSIP & 1 != 0).
*/
void OS_ISR_Software(void) {
  OS_INT_Enter();
  //
  // Perform any functionality here.
  //
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_ISR_Timer()
*
*  Function description
*    This routine performs the embOS tick handling.
*
*  Additional information
*    OS_ISR_Timer() is called when the Machine Timer interrupt is pending.
*    Machine Timer Interrupt becomes pending when (COREPLEX_MTIMECMP >= COREPLEX_MTIME).
*/
void OS_ISR_Timer(void) {
  OS_U64 Compare;

  OS_INT_Enter();
  Compare = _MTIME_GetCompare();  // Read time compare value
  do {                            // We might have to perform numerous ticks when interrupts have been disabled for extended periods
    OS_TICK_HandleEx();
    Compare += OS_TIMER_RELOAD;
  } while (Compare <= _MTIME_GetCounter());
  _MTIME_SetCompare(Compare);     // Eventually, configure new compare value. Clears pending bit as well.
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_TrapHandler()
*
*  Function description
*    This routine is used to handle traps routed to the mtvec vector.
*    It checks whether the trap was caused by an exception or an
*    interrupt and decides how to continue trap handling.
*
*  Additional information
*    Handling of traps highly depends on the implemented interrupt
*    controller and the used mode. Thus, the implementation of the
*    OS_TrapHandler() might me different for other risc-v devices.
*
*    OS_TrapHandler() forwards exceptions to OS_ExceptionHandler().
*
*    When CLINT/PLIC are used:
*    * In direct mode, this routine is called from the trap_entry(),
*      and is used for all interrupt requests of any kind. It detects
*      the specific reason for the local IRQ and calls the respective
*      interrupt service routine that was previously installed by
*      OS_RISCV_ISR_InstallHandler().
*    * In vectored mode, this routine is not used for interrupts, since
*      local interrupts are delegated through a ROM vector table instead.
*      The first entry in the vector table is still used as a trap handler
*      for exceptions and thus calls the OS_TrapHandler().
*
*    When ECLIC is used:
*      The ECLIC has multiple CSRs that can hold addresses for vectored
*      and non-vectored traps. The mnvec CSR can be used for NMIs. The mtvt
*      CSR is used for vectored interrupts. The mtvt2 CSR can be used for
*      non-vectored interrupts. The mtvec CSR can be used for all non-vectored
*      traps. By default, the BSP is set up to use mtvec for NMIs and exceptions,
*      mtvt for vectored interrupts and mtvt2 for non-vectored interrupts.
*/
void OS_TrapHandler(OS_U32 mcause, OS_U32 mepc) {
  if (mcause & MCAUSE_INT) {  // Trap caused by interrupt, call appropriate handler.
    //
    // This case should never happen as ECLIC is set up to
    // route exceptions and interrupts to separate trap vectors.
    //
    while (1);
  } else {                    // Trap caused by synchronous exception, call fault handler.
    OS_ExceptionHandler(mcause, mepc);
  }
}

/*********************************************************************
*
*       OS_ExceptionHandler()
*
*  Function description
*    This routine may be used to investigate synchronous traps.
*
*  Additional information
*    OS_ExceptionHandler() is called when a synchronous trap has occurred.
*
*    When CLINT/PLIC are used:
*      In direct mode, this routine is called from OS_TrapHandler().
*      In vectored mode, this routine is called from the ROM vector table.
*
*    When ECLIC is used:
*      This routine is called by OS_TrapHandler().
*/
void OS_ExceptionHandler(OS_U32 mcause, OS_U32 mepc) {
  OS_USEPARA(mepc);                    // May be used for further investigation.

#if (OS_DEBUG != 0)
  switch (mcause & ECLIC_EXCEPTION) {  // Detect specific exception cause:
  case  0: while (1);                  // - Instruction address misaligned.
  case  1: while (1);                  // - Instruction access fault.
  case  2: while (1);                  // - Illegal instruction.
  case  3: while (1);                  // - Breakpoint.
  case  4: while (1);                  // - Load address misaligned.
  case  5: while (1);                  // - Load access fault.
  case  6: while (1);                  // - Store/AMO address misaligned.
  case  7: while (1);                  // - Store/AMO access fault.
  case  8: while (1);                  // - Environment call from U-mode.
  case 11: while (1);                  // - Environment call from M-mode.
  default: while (1);                  // - Reserved. This should never happen.
  }
#else
  OS_USEPARA(mcause);
  while (1);  // Halt execution
#endif
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Set the ID for this hart
  //
  _OS_RISCV_HartID = OS_RISCV_GetHartID();
  //
  // Configure OS timer and enable Machine Timer Interrupt
  //
  _MTIME_SetCounter(0);
  _MTIME_SetCompare(OS_TIMER_RELOAD);
  OS_RISCV_ECLIC_ISR_Init((void*)ECLIC_BASE_ADDR, (void*)&vectors, (void*)&trap_entry);
  OS_RISCV_ECLIC_ISR_SetPriority(IRQ_M_TIMER, 0);
  OS_RISCV_ECLIC_ISR_Enable(IRQ_M_TIMER);
  //
  // Inform embOS about the timer settings
  //
  SystemCoreClockUpdate();
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  OS_TICK_Config(OS_TICK_FREQ, OS_INT_FREQ);
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_PROFILE != 0)
   SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {            // Idle loop: No task is ready to execute
  while (1) {                   // Nothing to do ... wait for interrupt
    #if (OS_DEBUG == 0)
      // When uncommenting this line, please be aware device
      // specific issues could occur.
      // Therefore, we do not call __WFI() by default.
      //
      //__asm volatile("wfi");  // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKMEM_SendChar(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
