/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2021 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.12.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : entry.s
Purpose : Entry point and vector table for the SiFive E31.
*/

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#if (__riscv_xlen == 64)
  #define STORE     sd
  #define LOAD      ld
  #define REGBYTES   8
#else
  #define STORE     sw
  #define LOAD      lw
  #define REGBYTES   4
#endif

/*********************************************************************
*
*       MACROS
*
**********************************************************************
*/

    .macro TRAP_ENTRY
        addi    sp, sp, -20 * REGBYTES
        //
        // Store caller save registers on the stack.
        // Callee save registers will be saved by the respective ISR function.
        //
        STORE   ra,  0 * REGBYTES(sp)
        STORE   t0,  1 * REGBYTES(sp)
        STORE   t1,  2 * REGBYTES(sp)
        STORE   t2,  3 * REGBYTES(sp)
        STORE   a0,  4 * REGBYTES(sp)
        STORE   a1,  5 * REGBYTES(sp)
        STORE   a2,  6 * REGBYTES(sp)
        STORE   a3,  7 * REGBYTES(sp)
        STORE   a4,  8 * REGBYTES(sp)
        STORE   a5,  9 * REGBYTES(sp)
        STORE   a6, 10 * REGBYTES(sp)
        STORE   a7, 11 * REGBYTES(sp)
        STORE   t3, 12 * REGBYTES(sp)
        STORE   t4, 13 * REGBYTES(sp)
        STORE   t5, 14 * REGBYTES(sp)
        STORE   t6, 15 * REGBYTES(sp)
        //
        // Store the mepc on the stack.
        // This is needed for nestable interrupts.
        //
        csrr    a1, mepc
        STORE   a1, 16 * REGBYTES(sp)
    .endm

    .macro TRAP_EXIT
        //
        // Load the mepc from the stack into the mepc register.
        // This is needed for nestable interrupts.
        //
        LOAD    a1, 16 * REGBYTES(sp)
        csrw    mepc, a1

        // Remain in M-mode after mret
        li      t0, 0x1800
        csrs    mstatus, t0
        //
        // Load caller save registers from stack and return.
        //
        LOAD    ra,  0 * REGBYTES(sp)
        LOAD    t0,  1 * REGBYTES(sp)
        LOAD    t1,  2 * REGBYTES(sp)
        LOAD    t2,  3 * REGBYTES(sp)
        LOAD    a0,  4 * REGBYTES(sp)
        LOAD    a1,  5 * REGBYTES(sp)
        LOAD    a2,  6 * REGBYTES(sp)
        LOAD    a3,  7 * REGBYTES(sp)
        LOAD    a4,  8 * REGBYTES(sp)
        LOAD    a5,  9 * REGBYTES(sp)
        LOAD    a6, 10 * REGBYTES(sp)
        LOAD    a7, 11 * REGBYTES(sp)
        LOAD    t3, 12 * REGBYTES(sp)
        LOAD    t4, 13 * REGBYTES(sp)
        LOAD    t5, 14 * REGBYTES(sp)
        LOAD    t6, 15 * REGBYTES(sp)
        addi    sp, sp, 20 * REGBYTES
        mret
    .endm

    .macro WEAK_DECL Name
        .weak \Name
        .type \Name, function
\Name\():
    .endm

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       Reset_Handler
*/
        .global Reset_Handler
        .section .app_entry, "ax", %progbits
        .balign 4
Reset_Handler:
//
// Jump to program start
//
        .option push
        .option norelax
        la      a0, _start
        jr      a0
        .option pop

/*********************************************************************
*
*       Vector table
*/
        .globl vtrap_entry
        .section .vectors, "ax", %progbits
        .balign 256
        .option push
        .option norvc
vtrap_entry:
        j       trap_entry
        j       .  // reserved
        j       .  // reserved
        j       MSI_Handler
        j       .  // reserved
        j       .  // reserved
        j       .  // reserved
        j       MTI_Handler
        j       .  // reserved
        j       .  // reserved
        j       .  // reserved
        j       MEI_Handler
        j       .  // reserved
        j       .  // reserved
        j       .  // reserved
        j       .  // reserved
        j       LI0_Handler
        j       LI1_Handler
        j       LI2_Handler
        j       LI3_Handler
        j       LI4_Handler
        j       LI5_Handler
        j       LI6_Handler
        j       LI7_Handler
        j       LI8_Handler
        j       LI9_Handler
        j       LI10_Handler
        j       LI11_Handler
        j       LI12_Handler
        j       LI13_Handler
        j       LI14_Handler
        j       LI15_Handler
        .option pop

/*********************************************************************
*
*       MSI_Handler
*/
MSI_Handler:
        TRAP_ENTRY
        call    OS_ISR_Software  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       MTI_Handler
*/
MTI_Handler:
        TRAP_ENTRY
        call    OS_ISR_Timer  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       MEI_Handler
*/
MEI_Handler:
        TRAP_ENTRY
        call    OS_ISR_External  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       LI0_Handler
*/
LI0_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local0
        TRAP_EXIT

/*********************************************************************
*
*       LI1_Handler
*/
LI1_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local1
        TRAP_EXIT

/*********************************************************************
*
*       LI2_Handler
*/
LI2_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local2
        TRAP_EXIT

/*********************************************************************
*
*       LI3_Handler
*/
LI3_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local3
        TRAP_EXIT

/*********************************************************************
*
*       LI4_Handler
*/
LI4_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local4
        TRAP_EXIT

/*********************************************************************
*
*       LI5_Handler
*/
LI5_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local5
        TRAP_EXIT

/*********************************************************************
*
*       LI6_Handler
*/
LI6_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local6
        TRAP_EXIT

/*********************************************************************
*
*       LI7_Handler
*/
LI7_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local7
        TRAP_EXIT

/*********************************************************************
*
*       LI8_Handler
*/
LI8_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local8
        TRAP_EXIT

/*********************************************************************
*
*       LI9_Handler
*/
LI9_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local9
        TRAP_EXIT

/*********************************************************************
*
*       LI10_Handler
*/
LI10_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local10
        TRAP_EXIT

/*********************************************************************
*
*       LI11_Handler
*/
LI11_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local11
        TRAP_EXIT

/*********************************************************************
*
*       LI12_Handler
*/
LI12_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local12
        TRAP_EXIT

/*********************************************************************
*
*       LI13_Handler
*/
LI13_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local13
        TRAP_EXIT

/*********************************************************************
*
*       LI14_Handler
*/
LI14_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local14
        TRAP_EXIT

/*********************************************************************
*
*       LI15_Handler
*/
LI15_Handler:
        TRAP_ENTRY
        call    OS_ISR_Local15
        TRAP_EXIT

/*** Weak high-level IRQ handler, defaulting to OS_ISR_Undefined(). Can be overwritten in "C" ***/
        WEAK_DECL OS_TrapHandler
        WEAK_DECL OS_ISR_Software
        WEAK_DECL OS_ISR_Timer
        WEAK_DECL OS_ISR_External
        WEAK_DECL OS_ISR_Local0
        WEAK_DECL OS_ISR_Local1
        WEAK_DECL OS_ISR_Local2
        WEAK_DECL OS_ISR_Local3
        WEAK_DECL OS_ISR_Local4
        WEAK_DECL OS_ISR_Local5
        WEAK_DECL OS_ISR_Local6
        WEAK_DECL OS_ISR_Local7
        WEAK_DECL OS_ISR_Local8
        WEAK_DECL OS_ISR_Local9
        WEAK_DECL OS_ISR_Local10
        WEAK_DECL OS_ISR_Local11
        WEAK_DECL OS_ISR_Local12
        WEAK_DECL OS_ISR_Local13
        WEAK_DECL OS_ISR_Local14
        WEAK_DECL OS_ISR_Local15
        call    OS_ISR_Undefined  // Implemented in RTOSInit.c
        ret

/********************************************************************/

        .end

/*************************** End of file ****************************/
