/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2021 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.12.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   USE_VECTORED_INT_MODE       // Modification to this value also requires modification to riscv_crt0.s! Consider using project options instead.
  #define USE_VECTORED_INT_MODE  (1)  // Configures ISR handling for vectored mode, in which a ROM vector table will be used for local interrupts.
#endif

/*********************************************************************
*
*       System tick settings
*/

#define OS_TIMER_FREQ    (32768u)
#define OS_TICK_FREQ     ( 1000u)
#define OS_INT_FREQ      ( 1024u)

#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
//
//  Coreplex-Local Interrupts
//
#define COREPLEX_BASE_ADDR             (0x02000000u)
#define COREPLEX_MSIP                  (*(volatile OS_U32*)(COREPLEX_BASE_ADDR + 0x0000u + (8u * _OS_RISCV_HartID)))  // LSB of this register may be used to generate SW interrupts
#define COREPLEX_MTIMECMP              (*(volatile OS_U64*)(COREPLEX_BASE_ADDR + 0x4000u + (8u * _OS_RISCV_HartID)))  // Used to generate OS tick, Timer compare register
#define COREPLEX_MTIME                 (*(volatile OS_U64*)(COREPLEX_BASE_ADDR + 0xBFF8u))                            // Used to generate OS tick, Timer counter register

#define COREPLEX_LOCAL_NUM_INTERRUPTS  (32u)

//
// Coreplex Platform-Level Interrupt Controller
//
#define COREPLEX_PLIC_BASE_ADDR        (0x0C000000u)

#define COREPLEX_PLIC_NUM_INTERRUPTS   (52u)
#define COREPLEX_PLIC_NUM_INT_PRIOS    ( 7u)

/*********************************************************************
*
*       Externals
*
**********************************************************************
*/
extern unsigned int vtrap_entry;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_U32 _OS_RISCV_HartID = 0u;

#if (USE_VECTORED_INT_MODE == 0)
static OS_ISR_HANDLER* _apLocalIsrHandler[COREPLEX_LOCAL_NUM_INTERRUPTS]   __attribute__ ((aligned (4), section (".non_init")));
#endif
static OS_ISR_HANDLER* _apExternalIsrHandler[COREPLEX_PLIC_NUM_INTERRUPTS] __attribute__ ((aligned (4), section (".non_init")));

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  unsigned int Result;
  OS_I64       Diff;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = COREPLEX_MTIMECMP;
  Counter = COREPLEX_MTIME;
  Diff    = (OS_I64)(Counter - Compare);
  if (Diff < 0) {
    Result = (unsigned int)(OS_TIMER_RELOAD + Diff);
  }  else {
    Result = (unsigned int)(Diff);
  }
  return Result;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  unsigned int IsPending;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = COREPLEX_MTIMECMP;
  Counter = COREPLEX_MTIME;
  if (Compare <= Counter) {
    IsPending = 1u;
  } else {
    IsPending = 0u;
  }
  return IsPending;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_ISR_Undefined()
*
*  Function description
*    This routine may be used to detect non-installed interrupts.
*
*  Additional information
*    OS_ISR_Undefined() is called when an interrupt is pending for
*    which no specific interrupt handler was previously installed.
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_Undefined(void) {
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    //
    // You may set a breakpoint here to detect non-installed ISR-handler
    //
  }
}

/*********************************************************************
*
*       OS_ISR_External()
*
*  Function description
*    This routine detects the specific reason for the external IRQ and
*    calls the respective interrupt service routine that was previously
*    installed by OS_RISCV_COREPLEX_ISR_InstallHandler().
*
*  Additional information
*    OS_ISR_External() is called when the Machine External interrupt is
*    pending. Machine External interrupt becomes pending when it is issued
*    by the external Platform-Level interrupt controller (PLIC).
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_External(void) {
  OS_U32 ISRIndex;

  ISRIndex = OS_RISCV_COREPLEX_ISR_Claim();         // Claim external interrupt by reading its ID, also clears pending bit
  if (ISRIndex != 0u) {                             // "0" indicates no IRQ, e.g. when it was already claimed by another hart
    OS_INT_Enter();
    _apExternalIsrHandler[ISRIndex]();              // Call appropriate handler
    OS_RISCV_COREPLEX_ISR_Complete(ISRIndex);       // Signal ISR completion
    OS_INT_Leave();
  }
}

/*********************************************************************
*
*       OS_ISR_Software()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    OS_ISR_Software() is called when the Machine Software interrupt is pending.
*    Machine Software interrupt becomes pending when (COREPLEX_MSIP & 1 != 0).
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_Software(void) {
  COREPLEX_MSIP &= ~1;     // Explicitly clear Machine Software interrupt pending bit.
  OS_INT_Enter();
  //
  // Perform any functionality here.
  //
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_ISR_Timer()
*
*  Function description
*    This routine performs the embOS tick handling.
*
*  Additional information
*    OS_ISR_Timer() is called when the Machine Timer interrupt is pending.
*    Machine Timer Interrupt becomes pending when (COREPLEX_MTIMECMP >= COREPLEX_MTIME).
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_Timer(void) {
  OS_U64 Compare;

  OS_INT_Enter();
  Compare = COREPLEX_MTIMECMP;  // Read time compare value
  do {                          // We might have to perform numerous ticks when interrupts have been disabled for extended periods
    OS_TICK_HandleEx();
    Compare    += OS_TIMER_RELOAD;
  } while (Compare <= COREPLEX_MTIME);
  COREPLEX_MTIMECMP = Compare;  // Eventually, configure new compare value. Clears pending bit as well.
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_TrapHandler()
*
*  Function description
*    This routine is used to handle traps routed to the mtvec vector.
*    It checks whether the trap was caused by an exception or an
*    interrupt and decides how to continue trap handling.
*
*  Additional information
*    Handling of traps highly depends on the implemented interrupt
*    controller and the used mode. Thus, the implementation of the
*    OS_TrapHandler() might me different for other risc-v devices.
*
*    OS_TrapHandler() forwards exceptions to OS_ExceptionHandler().
*
*    When CLINT/PLIC are used:
*    * In direct mode, this routine is called from the trap_entry(),
*      and is used for all interrupt requests of any kind. It detects
*      the specific reason for the local IRQ and calls the respective
*      interrupt service routine that was previously installed by
*      OS_RISCV_ISR_InstallHandler().
*    * In vectored mode, this routine is not used for interrupts, since
*      local interrupts are delegated through a ROM vector table instead.
*      The first entry in the vector table is still used as a trap handler
*      for exceptions and thus calls the OS_TrapHandler().
*
*    When ECLIC is used:
*      The ECLIC has multiple CSRs that can hold addresses for vectored
*      and non-vectored traps. The mnvec CSR can be used for NMIs. The mtvt
*      CSR is used for vectored interrupts. The mtvt2 CSR can be used for
*      non-vectored interrupts. The mtvec CSR can be used for all non-vectored
*      traps. By default, the BSP is set up to use mtvec for NMIs and exceptions,
*      mtvt for vectored interrupts and mtvt2 for non-vectored interrupts.
*/
void OS_TrapHandler(OS_U32 mcause, OS_U32 mepc) {
  if (mcause & MCAUSE_INT) {  // Trap caused by interrupt, call appropriate handler.
#if (USE_VECTORED_INT_MODE == 0)
    _apLocalIsrHandler[mcause & MCAUSE_CAUSE]();
#endif
  } else {                    // Trap caused by synchronous exception, call fault handler.
    OS_ExceptionHandler(mcause, mepc);
  }
}

/*********************************************************************
*
*       OS_ExceptionHandler()
*
*  Function description
*    This routine may be used to investigate synchronous traps.
*
*  Additional information
*    OS_ExceptionHandler() is called when a synchronous trap has occurred.
*
*    When CLINT/PLIC are used:
*    In direct mode, this routine is called from OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*
*    When ECLIC is used:
*      This routine is called by OS_TrapHandler().
*/
void OS_ExceptionHandler(OS_U32 mcause, OS_U32 mepc) {
  OS_USEPARA(mepc);                 // May be used for further investigation.

#if (OS_DEBUG != 0)
  switch (mcause & MCAUSE_CAUSE) {  // Detect specific exception cause:
  case  0: while (1);               // - Instruction address misaligned.
  case  1: while (1);               // - Instruction access fault.
  case  2: while (1);               // - Illegal instruction.
  case  3: while (1);               // - Breakpoint.
  case  4: while (1);               // - Load address misaligned.
  case  5: while (1);               // - Load access fault.
  case  6: while (1);               // - Store/AMO address misaligned.
  case  7: while (1);               // - Store/AMO access fault.
  case  8: while (1);               // - Environment call from U-mode.
  case 11: while (1);               // - Environment call from M-mode.
  default: while (1);               // - Reserved. This should never happen.
  }
#else
  OS_USEPARA(mcause);
  while (1);  // Halt execution
#endif
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_U32 i;

  OS_INT_IncDI();
  //
  // Set the ID for this hart
  //
  _OS_RISCV_HartID = OS_RISCV_GetHartID();
  //
  // ***** Initialize RISC-V interrupt handling, local *****
  //
#if (USE_VECTORED_INT_MODE == 0)
  //
  // Initialize direct ISR handling: install dummy handlers to detect non-installed interrupts
  //
  OS_RISCV_ISR_Init(COREPLEX_LOCAL_NUM_INTERRUPTS, _apLocalIsrHandler);  // Implicitly disables all sources
  for (i = 0u; i < COREPLEX_LOCAL_NUM_INTERRUPTS; i++) {
    OS_RISCV_ISR_InstallHandler(i, OS_ISR_Undefined);              // When in direct mode, install dummy handler to detect non-installed local interrupts
  }
#else
  //
  // Replace the default trap_entry which was set during startup
  // with the vector table and set the mode to vectored.
  //
  __asm volatile ("csrw mtvec, %0 \n\t"
                  :                                // Output list
                  : "r" ((int)&vtrap_entry | 0x1)  // Input list
                  : );                             // Clobbered list
  //
  // Initialize vectored ISR handling: initially disable all sources (can't assume these are reset)
  //
  OS_RISCV_ISR_Disable(IRQ_M_SOFTWARE);                            // Disable Machine Software Interrupt
  OS_RISCV_ISR_Disable(IRQ_M_TIMER);                               // Disable Machine Timer Interrupt
  OS_RISCV_ISR_Disable(IRQ_M_EXTERNAL);                            // Disable Machine External Interrupt
  for (i = IRQ_LOCAL0; i < COREPLEX_LOCAL_NUM_INTERRUPTS; i++) {
    OS_RISCV_ISR_Disable(i);                                       // Disable local interrupt 0 to 15 (if available)
  }
#endif
  //
  // Configure OS timer and enable Machine Timer Interrupt
  //
  COREPLEX_MTIME    = 0;                                           // Configure counter register
  COREPLEX_MTIMECMP = OS_TIMER_RELOAD;                             // Configure compare register
#if (USE_VECTORED_INT_MODE == 0)
  OS_RISCV_ISR_InstallHandler(IRQ_M_TIMER, OS_ISR_Timer);          // When in direct mode, install handler for Machine Timer Interrupt
#endif
  OS_RISCV_ISR_Enable(IRQ_M_TIMER);                                // Enable Machine Timer Interrupt
  //
  // Install and enable some additional local interrupts (for demonstration on how to use local interrupt sources)
  //
#if (USE_VECTORED_INT_MODE == 0)
  OS_RISCV_ISR_InstallHandler(IRQ_M_SOFTWARE, OS_ISR_Software);    // When in direct mode, install handler for Machine Software Interrupt
  OS_RISCV_ISR_InstallHandler(IRQ_M_EXTERNAL, OS_ISR_External);    // When in direct mode, install handler for Machine External Interrupt
#endif
  OS_RISCV_ISR_Enable(IRQ_M_SOFTWARE);                             // Enable Machine Software Interrupt
  OS_RISCV_ISR_Enable(IRQ_M_EXTERNAL);                             // Enable Machine External Interrupt
  //
  // ***** Initialize RISC-V interrupt handling, external *****
  //
  //
  // Install dummy handler to detect non-installed interrupts
  //
  OS_RISCV_COREPLEX_ISR_Init(COREPLEX_PLIC_BASE_ADDR, COREPLEX_PLIC_NUM_INTERRUPTS, COREPLEX_PLIC_NUM_INT_PRIOS, _apExternalIsrHandler);  // Implicitly disables all sources
  for (i = 0u; i < COREPLEX_PLIC_NUM_INTERRUPTS; i++) {
    OS_RISCV_COREPLEX_ISR_InstallHandler(i, OS_ISR_Undefined);     // Install dummy handler to detect non-installed external interrupts
  }
  //
  // Install and enable one external interrupt (for demonstration on how to use external interrupt sources)
  //
  //
  // Inform embOS about the timer settings
  //
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  OS_TICK_Config(OS_TICK_FREQ, OS_INT_FREQ);
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_PROFILE != 0)
  SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {            // Idle loop: No task is ready to execute
  while (1) {                   // Nothing to do ... wait for interrupt
    #if (OS_DEBUG == 0)
      // When uncommenting this line, please be aware device
      // specific issues could occur.
      // Therefore, we do not call __WFI() by default.
      //
      //__asm volatile("wfi");  // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKMEM_SendChar(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
